#import "StupidModelSwarm.h"

@implementation StupidModelSwarm

+ create: (id) aZone 
{
  StupidModelSwarm* obj = [super create: aZone];
  id <ProbeMap> probeMap;

  obj->cellInitialFoodAvailability = 0.0;
  obj->cellMaxFoodProductionRate = 0.01;
  obj->numberOfInitialBugs = 100;
  obj->initialBugSizeMean = 0.1;
  obj->initialBugSizeSD = 0.03;
  obj->bugMaxConsumptionRate = 1.0;
  obj->bugSurvivalProbability = 0.95;
  obj->bugSizeAverager = nil;
  obj->bsFilePtr = NULL;

  //
  // Starting in Version 5, the create method contains
  // the statements to build a probe map for the
  // parameters users can modify at run time
  //
  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "numberOfInitialBugs"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "bugMaxConsumptionRate"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "cellMaxFoodProductionRate"
				                inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "initialBugSizeMean"
				                inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "initialBugSizeSD"
				                inClass: [self class]]];
  	
   //
   // Now, install our custom probeMap class directly into the
   // probeLibrary
   //
   [probeLibrary setProbeMap: probeMap For: [self class]];


   return obj;
}



////////////////////////////////////
//
// buildObjects
//
///////////////////////////////////
- buildObjects 
{
  int i;
  int j;

  //fprintf(stdout, "StupidModelSwarm >>>> buildObject >>>> BEGIN\n");
  //fflush(0);


  //
  // allow our parent class to build anything.
  //
  [super buildObjects];

  //
  // Create a modelZone in which to manage our objects
  //
  modelZone = [Zone create: [self getZone]];

  //
  // Now set up the grid used to represent agent position
  //
  cellWorld = [Grid2d create: modelZone setSizeX: 100 Y: 100];


  //
  // Create a list for the cells
  //
  cellList = [List create: modelZone]; 
  
  //
  // Create the cells, so the bugs can move into them
  //
  for(i=0; i<100; i++)
  {
      for(j=0; j<100; j++)
      {
          HabitatCell* newCell = [HabitatCell create: modelZone];
          [newCell setStupidModelSwarm: self];
          [newCell setMyX: i];
          [newCell setMyY: j];
          [newCell setCellFoodAvailable: cellInitialFoodAvailability];
          [cellWorld putObject: newCell atX: i Y: j];
          [cellList addLast: newCell];
      }
  }

  // Now set up the separate grid used for bug locations
  bugWorld = [Grid2d create: modelZone setSizeX: 100 Y: 100];

  //
  // Create a list of the stupidBugs
  //
  bugList = [List create: modelZone];
  newBugList = [List create: modelZone];
  deadBugList = [List create: modelZone];
  

  //
  // For version 14, we create a normal distribution for
  // initial bug sizes here, and use it when creating initial bugs
  //
  bugInitSizeDist = [NormalDist create: modelZone 
                          setGenerator: randomGenerator
                               setMean: initialBugSizeMean
                             setStdDev: initialBugSizeSD];

  //
  // Make some agents
  // The bugs put themselves on the space.
  //
  for(i = 0; i < numberOfInitialBugs; i++) 
  {
      StupidBug* stupidBug = nil;
      double initBugSize = [bugInitSizeDist getDoubleSample];

      if(initBugSize <= 0.0) 
      {
          initBugSize = 0.0;
      }

      stupidBug = [StupidBug create: modelZone];
      [stupidBug setStupidModelSwarm: self];
      [stupidBug setBugWorld: bugWorld];
      [stupidBug setCellWorld: cellWorld];
      [stupidBug setRandPosition];
      [stupidBug setBugInitialSize: initBugSize];
 
      [bugList addLast: stupidBug];
  }

  // To be tidy, we drop the normal distribution
  [bugInitSizeDist drop];

  //
  // Version 8:
  // Create an Averager to get the statistics on bug
  // size written to an output file
  //
  bugSizeAverager = [Averager createBegin: modelZone];
  [bugSizeAverager setCollection: bugList];
  [bugSizeAverager setProbedSelector: M(getMySize)];
  bugSizeAverager = [bugSizeAverager createEnd];

  // Version 8: we need an output file and use standard io
  // Open the output file so that it overwrites if it already exists
 
  if((bsFilePtr = fopen("Stupid_output.txt", "w")) == NULL)
  {
      fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildObjects >>>> cannot open Stupid_output.txt for writing\n");
      fflush(0);
      exit(1);
  }

  //
  // Write a header
  //
  fprintf(bsFilePtr, "%-13s\t%-12s\t%-12s\t%-12s\n", "CurrentTime", "minBugSize", "aveBugSize", "maxBugSize");
  fflush(bsFilePtr);

  //fprintf(stdout, "StupidModelSwarm >>>> buildObject >>>> END\n");
  //fflush(0);

  return self;
}
	

 

/////////////////////////////////
//
// buildActions
//
/////////////////////////////////
- buildActions 
{
  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> BEGIN\n");

  [super buildActions];

  //
  // In version 10 we must start each time step
  // by sorting the bugs by size. This action calls the sort method.
  //
  updateActions = [ActionGroup create: modelZone];
  [updateActions createActionTo: self message: M(sortBugs)];

  //
  // Now the habitat cell actions
  //
  cellActions = [ActionGroup create: modelZone];
  [cellActions createActionForEach: cellList message: M(growFood)];


  //
  // Create an action group for agent actions
  //
  bugActions = [ActionGroup create: modelZone];

  //
  // Next, create the "move" and "grow" actions for bugs
  // Include the StupidBug interface (i.e. StupidBug.h) in StupidModelSwarm.h
  //
  [bugActions createActionForEach: bugList message: M(move)];
  [bugActions createActionForEach: bugList message: M(grow)];
  [bugActions createActionForEach: bugList message: M(die)];

  //
  // For version 12, create an action to execute the
  // addAndRemoveBugs action. This should be done
  // before we write output, so the output reflects
  // the current time step's reproduction and mortality
  
  // We put both this new action and file output in
  // one "cleanup" action group.
  //
  // Create an action group for file IO
  //
  cleanupActions = [ActionGroup create: modelZone];
  [cleanupActions createActionTo: self message: M(addAndRemoveBugs)];
  [cleanupActions createActionTo: self message: M(outputBugSize)];

  //
  // Then we create a schedule and put the 
  // action group(s) on it. 	      
  //
  modelSchedule = [Schedule createBegin: modelZone];
  [modelSchedule setRepeatInterval: 1];
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: updateActions];
  [modelSchedule at: 0 createAction: cellActions];
  [modelSchedule at: 0 createAction: bugActions];
  [modelSchedule at: 0 createAction: cleanupActions];

  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> END\n");

  return self;
}




/////////////////////////////////////////////
//
// activateIn
//
////////////////////////////////////////////
- activateIn:  swarmContext 
{
  //
  // First, activate ourselves via the superclass
  // activateIn: method.  Just pass along the context: the
  // activity library does the right thing.
  //
  [super activateIn: swarmContext];

  //
  // Now activate our own schedule.
  //
  [modelSchedule activateIn: self];

  //
  // Finally, return our activity.
  //
  return [self getSwarmActivity];
}



//////////////////////////////////////////
//
// updateActions
// 
//////////////////////////////////////////
- sortBugs
{
    [QSort sortObjectsIn: bugList];
    return self;
}

	  

//////////////////////////////////////////
//
// saysToStop 
//
//////////////////////////////////////////
- (BOOL) saysToStop 
{
    BOOL stoppingCriteriaMet = NO;
    int numBugs = [bugList getCount];

    // "getCurrentTime()" is a Swarm "macro" that returns current
    // time step. See the "Macro Index" in the Swarm reference manual.
    // The first time is zero.

    // Test print for time step:
    // fprintf(stdout, "Time step: %-4d \n", getCurrentTime());


    if(getCurrentTime() >= 999) 
    {
        stoppingCriteriaMet = YES;
    }
    else if(numBugs == 0) 
    {
        stoppingCriteriaMet = YES;
    }

    return stoppingCriteriaMet;
}


////////////////////////////////////////////
//
// outputBugSize
//
////////////////////////////////////////////
- outputBugSize
{
   double minBugSize;
   double meanBugSize;
   double maxBugSize;
   
   if(bsFilePtr == NULL)
   {
       fprintf(stderr, "ERROR: StupidModelSwarm >>>> outputBugSize >>>> bsFilePtr is NULL\n");
       fflush(0);
       exit(1);
   }

   [bugSizeAverager update];
   meanBugSize = [bugSizeAverager getAverage];
   minBugSize = [bugSizeAverager getMin];
   maxBugSize = [bugSizeAverager getMax];

   fprintf(bsFilePtr, "%-13ld\t%-12f\t%-12f\t%-12f\n", getCurrentTime(), minBugSize, meanBugSize, maxBugSize);
   fflush(bsFilePtr);

   return self;  
}

- addBug: aBug
{
    [newBugList addLast: aBug];
    return self;
}


- removeBug: aBug
{
    [deadBugList addLast: aBug];
    return self;
}
	
//
// For version 12, this method re-builds the bug list
// adding new ones and removing dead ones.
// And we must empty the lists of new and dead bugs!!
// New ones must be added first because some new bugs
// could be on the list of dead ones and need to be removed.
// This method has to be put on the schedule!
//
- addAndRemoveBugs 
{
   id <ListIndex> newBugNdx = [newBugList listBegin: scratchZone];
   StupidBug* nextNewBug;

   id <ListIndex> deadBugNdx = [deadBugList listBegin: scratchZone];
   StupidBug* nextDeadBug;

   while(([newBugNdx getLoc] != End) && ((nextNewBug = [newBugNdx next]) != nil)) 
   {
       [bugList addLast: nextNewBug];
   }
   [newBugNdx drop];

   [newBugList removeAll];

   while(([deadBugNdx getLoc] != End) && ((nextDeadBug = [deadBugNdx next]) != nil))
   {
       [bugList remove: nextDeadBug];
   }

   [deadBugNdx drop];

   [deadBugList deleteAll];

   return self;
}



////////////////////////////////
//
// getModelZone
//
///////////////////////////////
- (id <Zone>) getModelZone
{
    return modelZone;
}


/////////////////////////////
//
// getCellWorld
//
////////////////////////////
- (id <Grid2d>) getCellWorld
{
   return cellWorld;
}

/////////////////////////////
//
// getBugWorld
//
////////////////////////////
- (id <Grid2d>) getBugWorld
{
   return bugWorld;
}


///////////////////////////////
//
// getBugList
//
//////////////////////////////
- (id <List>) getBugList
{
    return bugList;
}

///////////////////////////////
//
// getCellList
//
//////////////////////////////
- (id <List>) getCellList
{
    return cellList;
}

- (double) getCellInitialFoodAvailability
{
    return cellInitialFoodAvailability;
}

- (double) getCellMaxFoodProductionRate
{
    return cellMaxFoodProductionRate;
}


- (double) getBugMaxConsumptionRate
{
    return bugMaxConsumptionRate;
}
		    
- (double) getBugSurvivalProbability
{
    return bugSurvivalProbability;
}

@end
