#import "StupidModelSwarm.h"

@implementation StupidModelSwarm

+ create: (id) aZone 
{
  StupidModelSwarm* obj = [super create: aZone];
  id <ProbeMap> probeMap;

  obj->cellInitialFoodAvailability = 0.0;
  obj->cellMaxFoodProductionRate = 0.01;
  obj->numberOfInitialBugs = 100;
  obj->initialBugSizeMean = 0.1;
  obj->initialBugSizeSD = 0.03;
  obj->bugMaxConsumptionRate = 1.0;
  obj->bugSurvivalProbability = 0.95;
  obj->bugSizeAverager = nil;
  obj->bsFilePtr = NULL;
  obj->worldMaxX = -1;
  obj->worldMaxY = -1;
  obj->cellFileName = "Stupid_Cell.Data";

  //
  // Starting in Version 5, the create method contains
  // the statements to build a probe map for the
  // parameters users can modify at run time
  //
  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "numberOfInitialBugs"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "bugMaxConsumptionRate"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "cellMaxFoodProductionRate"
				                inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "initialBugSizeMean"
				                inClass: [self class]]];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "initialBugSizeSD"
				                inClass: [self class]]];

  	
   //
   // Now, install our custom probeMap class directly into the
   // probeLibrary
   //
   [probeLibrary setProbeMap: probeMap For: [self class]];


   return obj;
}



////////////////////////////////////
//
// buildObjects
//
///////////////////////////////////
- buildObjects 
{
  int i;


  //fprintf(stdout, "StupidModelSwarm >>>> buildObject >>>> BEGIN\n");
  //fflush(0);

  //
  // allow our parent class to build anything.
  //
  [super buildObjects];

  //
  // Create a modelZone in which to manage our objects
  //
  modelZone = [Zone create: [self getZone]];

  //
  // Calculate the worldMaxX and worldMaxY; 
  // create bugWorld; 
  // create cellWorld;
  // build the cells.
  //
  [self buildCells];


  //
  // Create a list of the stupidBugs
  //
  bugList = [List create: modelZone];
  newBugList = [List create: modelZone];
  deadBugList = [List create: modelZone];
  
  //
  // and a list of predators
  //
  predList = [List create: modelZone];

  //
  // For version 14, we create a normal distribution for
  // initial bug sizes here, and use it when creating initial bugs
  //
  bugInitSizeDist = [NormalDist create: modelZone 
                          setGenerator: randomGenerator
                               setMean: initialBugSizeMean
                             setStdDev: initialBugSizeSD];

  //
  // Make some agents
  // The bugs put themselves on the space.
  //
  for(i = 0; i < numberOfInitialBugs; i++) 
  {
      StupidBug* stupidBug = nil;
      double initBugSize = [bugInitSizeDist getDoubleSample];

      if(initBugSize <= 0.0) 
      {
          initBugSize = 0.0;
      }

      stupidBug = [StupidBug create: modelZone];
      [stupidBug setStupidModelSwarm: self];
      [stupidBug setBugWorld: bugWorld];
      [stupidBug setCellWorld: cellWorld];
      [stupidBug setRandPosition];
      [stupidBug setBugInitialSize: initBugSize];
 
      [bugList addLast: stupidBug];
  }

  //
  // To be tidy, we drop the normal distribution
  //
  [bugInitSizeDist drop];

  //		
  // For version 16, we create predators and put them on a list
  // and on a separate space.
  //
  for(i = 0; i < 200; i++) 
  {
      StupidPredator* stupidPred = [StupidPredator create: modelZone];
      [stupidPred setStupidModelSwarm: self];
      [stupidPred setBugWorld: bugWorld];
      [stupidPred setCellWorld: cellWorld];
      [stupidPred setPredWorld: predWorld];
      [stupidPred setRandPosition];
 
      [predList addLast: stupidPred];
  }

  //
  // Version 8:
  // Create an Averager to get the statistics on bug
  // size written to an output file
  //
  bugSizeAverager = [Averager createBegin: modelZone];
  [bugSizeAverager setCollection: bugList];
  [bugSizeAverager setProbedSelector: M(getMySize)];
  bugSizeAverager = [bugSizeAverager createEnd];

  // Version 8: we need an output file and use standard io
  // Open the output file so that it overwrites if it already exists
 
  if((bsFilePtr = fopen("Stupid_output.txt", "w")) == NULL)
  {
      fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildObjects >>>> cannot open Stupid_output.txt for writing\n");
      fflush(0);
      exit(1);
  }

  //
  // Write a header
  //
  fprintf(bsFilePtr, "%-13s\t%-12s\t%-12s\t%-12s\n", "CurrentTime", "minBugSize", "aveBugSize", "maxBugSize");
  fflush(bsFilePtr);

  //fprintf(stdout, "StupidModelSwarm >>>> buildObject >>>> END\n");
  //fflush(0);

  return self;
}


/////////////////////////////////////////////
//
// buildCells
//
///////////////////////////////////////////
- buildCells
{
    FILE* fptr = NULL;
    char inputString[200];    
    int nextX;
    int nextY;
    double foodProductionRate;
    int lineNo = 0;

    //fprintf(stdout, "StupidModelSwarm >>>> buildCells >>>> BEGIN\n");
    //fflush(0);

    if((fptr = fopen(cellFileName, "r")) == NULL)
    {
        fprintf(stdout, "ERROR: StupidModelSwarm >>>> buildCells >>>> unable to open %s for reading\n", cellFileName);
        fflush(0);
        exit(1);
    }

    //
    // skip the first 3 lines, since that's the format of data files
    //
    fgets(inputString, 200, fptr);
    fgets(inputString, 200, fptr);
    fgets(inputString, 200, fptr);

    //
    // Get the size of the world from the data file.
    //
    lineNo = 3;
    while(fgets(inputString, 200, fptr) != NULL)
    {
        lineNo++;
        if(sscanf(inputString, "%d %d %lf", &nextX, &nextY, &foodProductionRate) != 3)
        { 
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> incorrect number of arguments \n");
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> check lineNo %d in %s\n", lineNo, cellFileName);
            fflush(0);
            exit(1);
        }
        
        worldMaxX = (worldMaxX > nextX) ? worldMaxX : nextX; 
        worldMaxY = (worldMaxY > nextY) ? worldMaxY : nextY; 
    }

 
    //
    // Create a list for the cells
    //
    cellList = [List create: modelZone]; 
  

    //
    // Now set up the grid used to represent agent position
    //
    cellWorld = [Grid2d create: modelZone setSizeX: worldMaxX + 1 Y: worldMaxY + 1];

    //
    // Now set up the separate grid used for bug locations
    //
    bugWorld = [Grid2d create: modelZone setSizeX: worldMaxX + 1 Y: worldMaxY + 1];

    //
    // and a grid for the predators
    //
    predWorld = [Grid2d create: modelZone setSizeX: worldMaxX + 1 Y: worldMaxY + 1]; 


    //
    // Set the file pointer back to the beginning of the file.
    //
    rewind(fptr);

    //
    // Again, skip the first 3 lines, since that's the format of data files
    //
    fgets(inputString, 200, fptr);
    fgets(inputString, 200, fptr);
    fgets(inputString, 200, fptr);
    //
    // Create the cells, so the bugs can move into them
    //
    lineNo = 3;
    while(fgets(inputString, 200, fptr) != NULL)
    {
        HabitatCell* newCell = [HabitatCell create: modelZone];

        lineNo++;
        if(sscanf(inputString, "%d %d %lf", &nextX, &nextY, &foodProductionRate) != 3)
        { 
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> incorrect number of arguments \n");
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> check lineNo %d in %s\n", lineNo, cellFileName);
            fflush(0);
            exit(1);
        }

        [newCell setStupidModelSwarm: self];
        [newCell setMyX: nextX];
        [newCell setMyY: nextY];
        [newCell setCellFoodAvailable: cellInitialFoodAvailability];
        [newCell setCellFoodProductionRate: foodProductionRate];

        if([cellWorld getObjectAtX: nextX Y: nextY] != nil)
        {
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> cell at %d %d already exits\n", worldMaxX, worldMaxY);
            fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildCells >>>> check lineNo %d in %s\n", lineNo, cellFileName);
            fflush(0);
            exit(1);
        }

        [cellWorld putObject: newCell atX: nextX Y: nextY];

        [cellList addLast: newCell];
    }

    fclose(fptr);
         
    //fprintf(stdout, "StupidModelSwarm >>>> buildCells >>>> END\n");
    //fflush(0);

    return self;
}

/////////////////////////////////
//
// buildActions
//
/////////////////////////////////
- buildActions 
{
  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> BEGIN\n");

  [super buildActions];

  //
  // In version 10 we must start each time step
  // by sorting the bugs by size. This action calls the sort method.
  //
  updateActions = [ActionGroup create: modelZone];
  [updateActions createActionTo: self message: M(sortBugs)];

  //
  // Now the habitat cell actions
  //
  cellActions = [ActionGroup create: modelZone];
  [cellActions createActionForEach: cellList message: M(growFood)];


  //
  // Create an action group for bug actions
  //
  bugActions = [ActionGroup create: modelZone];

  //
  // Next, create the "move" and "grow" actions for bugs
  // Include the StupidBug interface (i.e. StupidBug.h) in StupidModelSwarm.h
  //
  [bugActions createActionForEach: bugList message: M(move)];
  [bugActions createActionForEach: bugList message: M(grow)];
  [bugActions createActionForEach: bugList message: M(die)];

  //
  // Version 16: Create an action group for predators
  // and put their "hunt" action on it.
  //
  predatorActions = [ActionGroup create: modelZone];
  [predatorActions createActionForEach: predList message: M(hunt)];

  //
  // For version 12, create an action to execute the
  // addAndRemoveBugs action. This should be done
  // before we write output, so the output reflects
  // the current time step's reproduction and mortality
  
  // We put both this new action and file output in
  // one "cleanup" action group.
  //
  // Create an action group for file IO
  //
  cleanupActions = [ActionGroup create: modelZone];
  [cleanupActions createActionTo: self message: M(addAndRemoveBugs)];
  [cleanupActions createActionTo: self message: M(outputBugSize)];

  //
  // Then we create a schedule and put the 
  // action group(s) on it. 	      
  //
  modelSchedule = [Schedule createBegin: modelZone];
  [modelSchedule setRepeatInterval: 1];
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: updateActions];
  [modelSchedule at: 0 createAction: cellActions];
  [modelSchedule at: 0 createAction: bugActions];
  [modelSchedule at: 0 createAction: predatorActions];
  [modelSchedule at: 0 createAction: cleanupActions];

  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> END\n");

  return self;
}




/////////////////////////////////////////////
//
// activateIn
//
////////////////////////////////////////////
- activateIn:  swarmContext 
{
  //
  // First, activate ourselves via the superclass
  // activateIn: method.  Just pass along the context: the
  // activity library does the right thing.
  //
  [super activateIn: swarmContext];

  //
  // Now activate our own schedule.
  //
  [modelSchedule activateIn: self];

  //
  // Finally, return our activity.
  //
  return [self getSwarmActivity];
}



//////////////////////////////////////////
//
// updateActions
// 
//////////////////////////////////////////
- sortBugs
{
    [QSort sortObjectsIn: bugList];
    return self;
}

	  

//////////////////////////////////////////
//
// saysToStop 
//
//////////////////////////////////////////
- (BOOL) saysToStop 
{
    BOOL stoppingCriteriaMet = NO;
    int numBugs = [bugList getCount];

    // "getCurrentTime()" is a Swarm "macro" that returns current
    // time step. See the "Macro Index" in the Swarm reference manual.
    // The first time is zero.

    // Test print for time step:
    // fprintf(stdout, "Time step: %-4d \n", getCurrentTime());


    if(getCurrentTime() >= 999) 
    {
        stoppingCriteriaMet = YES;
    }
    else if(numBugs == 0) 
    {
        stoppingCriteriaMet = YES;
    }

    return stoppingCriteriaMet;
}


////////////////////////////////////////////
//
// outputBugSize
//
////////////////////////////////////////////
- outputBugSize
{
   double minBugSize;
   double meanBugSize;
   double maxBugSize;
   
   if(bsFilePtr == NULL)
   {
       fprintf(stderr, "ERROR: StupidModelSwarm >>>> outputBugSize >>>> bsFilePtr is NULL\n");
       fflush(0);
       exit(1);
   }

   [bugSizeAverager update];
   meanBugSize = [bugSizeAverager getAverage];
   minBugSize = [bugSizeAverager getMin];
   maxBugSize = [bugSizeAverager getMax];

   fprintf(bsFilePtr, "%-13ld\t%-12f\t%-12f\t%-12f\n", getCurrentTime(), minBugSize, meanBugSize, maxBugSize);
   fflush(bsFilePtr);

   return self;  
}

- addBug: aBug
{
    [newBugList addLast: aBug];
    return self;
}


- removeBug: aBug
{
    [deadBugList addLast: aBug];
    return self;
}
	
//
// For version 12, this method re-builds the bug list
// adding new ones and removing dead ones.
// And we must empty the lists of new and dead bugs!!
// New ones must be added first because some new bugs
// could be on the list of dead ones and need to be removed.
// This method has to be put on the schedule!
//
- addAndRemoveBugs 
{
   id <ListIndex> newBugNdx = [newBugList listBegin: scratchZone];
   StupidBug* nextNewBug;

   id <ListIndex> deadBugNdx = [deadBugList listBegin: scratchZone];
   StupidBug* nextDeadBug;

   //fprintf(stderr, "StupidModelSwarm >>>> addAndRemoveBugs >>>> BEGIN\n");
   //fflush(0);
   
   while(([newBugNdx getLoc] != End) && ((nextNewBug = [newBugNdx next]) != nil)) 
   {
       [bugList addLast: nextNewBug];
   }
   [newBugNdx drop];

   [newBugList removeAll];

   while(([deadBugNdx getLoc] != End) && ((nextDeadBug = [deadBugNdx next]) != nil))
   {
       [bugList remove: nextDeadBug];
   }

   [deadBugNdx drop];

   [deadBugList deleteAll];

   //fprintf(stderr, "StupidModelSwarm >>>> addAndRemoveBugs >>>> END\n");
   //fflush(0);

   return self;
}



////////////////////////////////
//
// getModelZone
//
///////////////////////////////
- (id <Zone>) getModelZone
{
    return modelZone;
}


/////////////////////////////
//
// getCellWorld
//
////////////////////////////
- (id <Grid2d>) getCellWorld
{
   return cellWorld;
}

/////////////////////////////
//
// getBugWorld
//
////////////////////////////
- (id <Grid2d>) getBugWorld
{
   return bugWorld;
}


/////////////////////////////
//
// getPredWorld
//
////////////////////////////
- (id <Grid2d>) getPredWorld
{
    return predWorld;
}

///////////////////////////////
//
// getBugList
//
//////////////////////////////
- (id <List>) getBugList
{
    return bugList;
}

///////////////////////////////
//
// getCellList
//
//////////////////////////////
- (id <List>) getCellList
{
    return cellList;
}



//////////////////////////////
//
// getPredList
//
/////////////////////////////
- (id <List>) getPredList
{
    return predList;
}



/////////////////////////////////////////
//
// getCellInitialFoodAvailability
//
////////////////////////////////////////
- (double) getCellInitialFoodAvailability
{
    return cellInitialFoodAvailability;
}


/////////////////////////////////////////
//
// getBugMaxConsumptionRate
//
////////////////////////////////////////
- (double) getBugMaxConsumptionRate
{
    return bugMaxConsumptionRate;
}
		    
/////////////////////////////////////////
//
// getBugSurvivalProbability
//
////////////////////////////////////////
- (double) getBugSurvivalProbability
{
    return bugSurvivalProbability;
}

@end
