#import "StupidModelSwarm.h"

#import <stdio.h>



@implementation StupidModelSwarm

+ create: (id) aZone 
{
  StupidModelSwarm* obj = [super create: aZone];
  id <ProbeMap> probeMap;

  obj->cellInitialFoodAvailability = 0.0;
  obj->cellMaxFoodProductionRate = 0.01;
  obj->numberOfInitialBugs = 100;
  obj->bugInitSize = 1.0;
  obj->bugMaxConsumptionRate = 1.0;
  obj->bugSizeAverager = nil;
  obj->bsFilePtr = NULL;

  //
  // Starting in Version 5, the create method contains
  // the statements to build a probe map for the
  // parameters users can modify at run time
  //
  probeMap = [EmptyProbeMap createBegin: aZone];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "numberOfInitialBugs"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "bugMaxConsumptionRate"
				                inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "cellMaxFoodProductionRate"
				                inClass: [self class]]];

  	
   //
   // Now, install our custom probeMap class directly into the
   // probeLibrary
   //
   [probeLibrary setProbeMap: probeMap For: [self class]];


   return obj;
}



////////////////////////////////////
//
// buildObjects
//
///////////////////////////////////
- buildObjects 
{
  int i;
  int j;

  fprintf(stdout, "StupidModelSwarm >>>> buildObjects >>>> BEGIN\n");
  fflush(0);

  //
  // allow our parent class to build anything.
  //
  [super buildObjects];

  //
  // Create a modelZone in which to manage our objects
  //
  modelZone = [Zone create: [self getZone]];

  //
  // Now set up the grid used to represent agent position
  //
  cellWorld = [Grid2d create: modelZone setSizeX: 100 Y: 100];


  //
  // Create a list for the cells
  //
  cellList = [List create: modelZone]; 
  
  //
  // Create the cells, so the bugs can move into them
  //
  for(i=0; i<100; i++)
  {
      for(j=0; j<100; j++)
      {
          HabitatCell* newCell = [HabitatCell create: modelZone];
          [newCell setStupidModelSwarm: self];
          [newCell setMyX: i];
          [newCell setMyY: j];
          [newCell setCellFoodAvailable: cellInitialFoodAvailability];
          [cellWorld putObject: newCell atX: i Y: j];
          [cellList addLast: newCell];
      }
  }

  // Now set up the separate grid used for bug locations
  bugWorld = [Grid2d create: modelZone setSizeX: 100 Y: 100];

  //
  // Create a list of the stupidBugs
  //
  bugList = [List create: modelZone];
  
  //
  // Make some agents
  // The bugs put themselves on the space.
  //



  for(i = 0; i < numberOfInitialBugs; i++) 
  {
      StupidBug* stupidBug = nil;

      stupidBug = [StupidBug create: modelZone];
      [stupidBug setStupidModelSwarm: self];
      [stupidBug setBugWorld: bugWorld];
      [stupidBug setCellWorld: cellWorld];
      [stupidBug setRandPosition];
      [stupidBug setBugInitialSize: bugInitSize];
 
      [bugList addLast: stupidBug];

  }

  //
  // Version 8:
  // Create an Averager to get the statistics on bug
  // size written to an output file
  //
  bugSizeAverager = [Averager createBegin: modelZone];
  [bugSizeAverager setCollection: bugList];
  [bugSizeAverager setProbedSelector: M(getMySize)];
  bugSizeAverager = [bugSizeAverager createEnd];

  // Version 8: we need an output file and use standard io
  // Open the output file so that it overwrites if it already exists
 
  if((bsFilePtr = fopen("Stupid_output.txt", "w")) == NULL)
  {
      fprintf(stderr, "ERROR: StupidModelSwarm >>>> buildObjects >>>> cannot open Stupid_output.txt for writing\n");
      fflush(0);
      exit(1);
  }

  //
  // Write a header
  //
  fprintf(bsFilePtr, "%-13s\t%-12s\t%-12s\t%-12s\n", "CurrentTime", "minBugSize", "aveBugSize", "maxBugSize");
  fflush(bsFilePtr);

  fprintf(stdout, "StupidModelSwarm >>>> buildObjects >>>> END\n");
  fflush(0);

  return self;
}
	

 

/////////////////////////////////
//
// buildActions
//
/////////////////////////////////
- buildActions 
{
  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> BEGIN\n");

  [super buildActions];

  //
  // Now the first actions are for the habitat cells
  //
  cellActions = [ActionGroup create: modelZone];
  [cellActions createActionForEach: cellList message: M(growFood)];


  //
  // Create an action group for agent actions
  //
  bugActions = [ActionGroup create: modelZone];

  // Next, create the "move" and "grow" actions for bugs
  // Include the StupidBug interface (i.e. StupidBug.h) in StupidModelSwarm.h
  //
  // For Version 9, we add a statement randomizing the order in which
  // "move" is executed over the bug list. The actionForEach object returned by
  // "createActionsForEach" is captured in the local variable bugMoveActionForEach
  // and then sent a message changing its default order to random
  //
  bugMoveActionForEach = [bugActions createActionForEach: bugList message: M(move)];
  [bugMoveActionForEach setDefaultOrder: Randomized];

  [bugActions createActionForEach: bugList message: M(grow)];

  //
  // Create an action group for file IO
  //
  fileIOActions = [ActionGroup create: modelZone];
  [fileIOActions createActionTo: self message: M(outputBugSize)];


  //
  // Then we create a schedule and put the 
  // action group(s) on it. 	      
  //
  modelSchedule = [Schedule createBegin: modelZone];
  [modelSchedule setRepeatInterval: 1];
  modelSchedule = [modelSchedule createEnd];

  [modelSchedule at: 0 createAction: cellActions];
  [modelSchedule at: 0 createAction: bugActions];
  [modelSchedule at: 0 createAction: fileIOActions];

  //fprintf(stdout, "StupidModelSwarm >>>> buildActions >>>> END\n");

  return self;
}




/////////////////////////////////////////////
//
// activateIn
//
////////////////////////////////////////////
- activateIn:  swarmContext 
{
  //
  // First, activate ourselves via the superclass
  // activateIn: method.  Just pass along the context: the
  // activity library does the right thing.
  //
  [super activateIn: swarmContext];

  //
  // Now activate our own schedule.
  //
  [modelSchedule activateIn: self];

  //
  // Finally, return our activity.
  //
  return [self getSwarmActivity];
}

	  

//////////////////////////////////////////
//
// saysToStop 
//
//////////////////////////////////////////
- (BOOL) saysToStop 
{
    double maxBugSize = -1.0;

    BOOL stoppingCriteriaMet = NO;

 //   [bugSizeAverager update];   We do not need to update the averager because
 //                               it was already updated in "outputBugSize"

    maxBugSize = [bugSizeAverager getMax];

    if(maxBugSize > 100)
    {
       stoppingCriteriaMet = YES;
       fclose(bsFilePtr);          // Close the output file when we stop
    }

    return stoppingCriteriaMet;
}


////////////////////////////////////////////
//
// outputBugSize
//
////////////////////////////////////////////
- outputBugSize
{
   double minBugSize;
   double meanBugSize;
   double maxBugSize;
   
   if(bsFilePtr == NULL)
   {
       fprintf(stderr, "ERROR: StupidModelSwarm >>>> outputBugSize >>>> bsFilePtr is NULL\n");
       fflush(0);
       exit(1);
   }

   [bugSizeAverager update];
   meanBugSize = [bugSizeAverager getAverage];
   minBugSize = [bugSizeAverager getMin];
   maxBugSize = [bugSizeAverager getMax];

   fprintf(bsFilePtr, "%-13ld\t%-12f\t%-12f\t%-12f\n", getCurrentTime(), minBugSize, meanBugSize, maxBugSize);
   fflush(bsFilePtr);

   return self;  
}

/////////////////////////////
//
// getCellWorld
//
////////////////////////////
- (id <Grid2d>) getCellWorld
{
   return cellWorld;
}

/////////////////////////////
//
// getBugWorld
//
////////////////////////////
- (id <Grid2d>) getBugWorld
{
   return bugWorld;
}


///////////////////////////////
//
// getBugList
//
//////////////////////////////
- (id <List>) getBugList
{
    return bugList;
}

///////////////////////////////
//
// getCellList
//
//////////////////////////////
- (id <List>) getCellList
{
    return cellList;
}

- (double) getCellInitialFoodAvailability
{
    return cellInitialFoodAvailability;
}

- (double) getCellMaxFoodProductionRate
{
    return cellMaxFoodProductionRate;
}

- (double) getBugInitSize
{
    return bugInitSize;
}

- (double) getBugMaxConsumptionRate
{
    return bugMaxConsumptionRate;
}
		    

@end
